#!/usr/bin/env python3

"""
Thaicert worker for the ACT platform

"""

import argparse
import sys
import traceback
import act
import act.api
import configparser
import os

from typing import Text, Dict, List
from itertools import combinations
from logging import error

from act.workers.libs import worker
from act.api.helpers import handle_fact


WORKER_NAME = "thaicert"
THAICERT_URL = "https://apt.thaicert.or.th/cgi-bin/getcard.cgi?g=all&o=j"
STIX_VOCAB = "http://raw.githubusercontent.com/oasis-open/cti-stix2-json-schemas/stix2.1/schemas/sdos/identity.json"
COUNTRY_REGIONS = "https://raw.githubusercontent.com/lukes/ISO-3166-Countries-with-Regional-Codes/master/all/all.json"
THAICERT_TOOLS_URL = "https://apt.thaicert.or.th/cgi-bin/getcard.cgi?t=all&o=j"

def parseargs() -> argparse.ArgumentParser:
    """ Parse arguments"""
    parser = worker.parseargs("ThaiCERT worker")
    parser.add_argument('--url', dest="thaicert_url", default=THAICERT_URL,
                        help='ThaiCERT data URL')
    return parser


def process(client: act.api.Act, ta_cards: List, output_format: Text = "json") -> None:
    "Extract threat actor cards from ThaiCERT"

    for actor in ta_cards:
        if len(actor["names"]) > 1:
            aliases = list(set([alias["name"] for alias in actor["names"]]))
            for a1, a2 in combinations(aliases, 2):
                handle_fact(
                    client.fact("alias")
                    .bidirectional("threatActor", a1, "threatActor", a2),
                    output_format=output_format
                )

        if "operations" in actor:
            for operation in actor["operations"]:
                if len(operation["activity"].split("\n")[0].split()) < 4:
                    for ta in actor["actor"].split(","):
                        chain = act.api.fact.fact_chain(
                            client.fact("attributedTo")
                            .source("incident", "*")
                            .destination("campaign", operation["activity"].split("\n")[0]),
                            client.fact("attributedTo")
                            .source("incident", "*")
                            .destination("threatActor", ta)
                        )

                for fact in chain:
                    try:
                        handle_fact(fact, output_format=output_format)
                    except act.api.base.ValidationError as err:
                        warning("ValidationError while storing objects: %s" % err)


def add_countries(client: act.api.Act, ta_cards: List, countries: List, output_format: Text = "json") -> None:
    """
    Only submit country if ISO-3166 country
    """
    for actor in ta_cards:
        if "country" in actor: 
            for country in actor["country"]:
                if country.lower() in countries:
                    chain = act.api.fact.fact_chain(
                        client.fact("locatedIn")
                        .source("organization", "*")
                        .destination("country", country),
                        client.fact("attributedTo")
                        .source("threatActor", actor["actor"])
                        .destination("organization", "*")
                    )

                    for fact in chain:
                        handle_fact(fact, output_format=output_format)

        if "observed-countries" in actor:
            for country in actor["observed-countries"]:
                if country.lower() in countries:
                    for ta in actor["actor"].split(","):
                        chain = act.api.fact.fact_chain(
                            client.fact("locatedIn")
                            .source("organization", "*")
                            .destination("country", country),
                            client.fact("targets")
                            .source("incident", "*")
                            .destination("organization", "*"),
                            client.fact("attributedTo")
                            .source("incident", "*")
                            .destination("threatActor", ta)
                        )

                        for fact in chain:
                            handle_fact(fact, output_format=output_format)


def add_sectors(client: act.api.Act, ta_cards: List, vocab: List, output_format: Text = "json") -> None:
    """
    Only submit sectors if in STIX vocabulary
    """
    for actor in ta_cards:
        if "observed-sectors" in actor:
            for sector in actor["observed-sectors"]:
                if sector.lower() in vocab["definitions"]["industry-sector-ov"]["enum"]:
                    for ta in actor["actor"].split(","):
                        chain = act.api.fact.fact_chain(
                            client.fact("memberOf")
                            .source("organization", "*")
                            .destination("sector", sector.lower()),
                            client.fact("targets")
                            .source("incident", "*")
                            .destination("organization", "*"),
                            client.fact("attributedTo")
                            .source("incident", "*")
                            .destination("threatActor", ta)
                        )

                    for fact in chain:
                        handle_fact(fact, output_format=output_format)


def add_tools(client: act.api.Act, ta_cards: List, tools: List, output_format: Text = "json") -> None:
    """
    Submit tool aliases and actor tools
    """
    tool_vocab = [tool["tool"] for tool in tools]
    for actor in ta_cards:
       if "tools" in actor:
            for tool in actor["tools"]:
                if tool in tool_vocab:
                    for ta in actor["actor"].split(","):
                        chain = act.api.fact.fact_chain(
                            client.fact("classifiedAs")
                            .source("content", "*")
                            .destination("tool", tool.lower()),
                            client.fact("observedIn")
                            .source("content", "*")
                            .destination("event", "*"),
                            client.fact("attributedTo")
                            .source("event", "*")
                            .destination("incident", "*"),
                            client.fact("attributedTo")
                            .source("incident", "*")
                            .destination("threatActor", ta)
                        )

                    for fact in chain:
                        try:
                            handle_fact(fact, output_format=output_format)
                        except act.api.base.ValidationError as err:
                            error("ResponseError while storing objects: %s" % err)

    for values in tools:
        aliases = set([tool["name"].strip() for tool in values["names"]] + [values["tool"]])
        for tool1, tool2 in combinations(aliases, 2):
            fact = client.fact("alias") \
                                .bidirectional("tool", tool1, "tool", tool2)
            handle_fact(fact)


def main() -> None:
    """Main function"""
    args = worker.handle_args(parseargs())
    actapi = worker.init_act(args)
    ta_cards = worker.fetch_json(args.thaicert_url, args.proxy_string, args.http_timeout)
    process(actapi, ta_cards["values"])

    vocab = worker.fetch_json(STIX_VOCAB, args.proxy_string, args.http_timeout)
    add_sectors(actapi, ta_cards["values"], vocab)

    countries = worker.fetch_json(COUNTRY_REGIONS, args.proxy_string, args.http_timeout)
    countries = [country["name"].lower() for country in countries]
    add_countries(actapi, ta_cards["values"], countries)

    tools = worker.fetch_json(THAICERT_TOOLS_URL, args.proxy_string, args.http_timeout)
    add_tools(actapi, ta_cards["values"], tools["values"])

def main_log_error() -> None:
    "Main function wrapper. Log all exceptions to error"
    try:
       main() 
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == '__main__':
    main_log_error()



""" Cache hash of content downloaded from URL """

import datetime
import hashlib
import os
import sqlite3
import time
from logging import info
from typing import Dict, NamedTuple, Optional, Text

import caep
import requests

from act.workers.libs import worker


class CacheResponse(NamedTuple):
    """ Response object from cache query """
    status_code: Optional[int]
    report_hash: Optional[Text]
    added: datetime.datetime


class UrlContentHash(NamedTuple):
    """ Response object with status code and sha256 hash """
    status_code: Optional[int]
    sha256: Optional[Text]


CACHE_PREFIX = "act-workers"


def get_db_cache(cache_dir: Text) -> sqlite3.Connection:
    """
    Open cache and return sqlite3 connection
    Table is created if it does not exists
    """
    cache_file = os.path.join(cache_dir, "urlcache.sqlite3")
    conn = sqlite3.connect(cache_file)
    cursor = conn.cursor()
    cursor.execute("""CREATE TABLE IF NOT EXISTS report_hash (
        url string primary key,
        status_code int,
        sha256 string,
        added int)
    """)
    cursor.execute("CREATE UNIQUE INDEX IF NOT EXISTS report_url on report_hash(url)")

    return conn


class URLCache:
    """ Cache hash of content downloaded from URL """

    def __init__(
            self,
            cache_prefix: Text = CACHE_PREFIX,
            requests_common_kwargs: Optional[Dict] = None) -> None:

        self.requests_common_kwargs = \
            requests_common_kwargs if requests_common_kwargs else {}

        cache_dir = caep.get_cache_dir(cache_prefix, create=True)
        self.cache: sqlite3.Connection = get_db_cache(cache_dir)

    def query_cache(self, url: Text) -> CacheResponse:
        """ Query cache for a specific url """
        cursor = self.cache.cursor()

        res = cursor.execute("SELECT * FROM report_hash WHERE url = ?", [url.strip()]).fetchall()

        if not res:
            return CacheResponse(None, None, datetime.datetime.utcfromtimestamp(0))

        return CacheResponse(res[0][1], res[0][2], datetime.datetime.utcfromtimestamp(res[0][3]))

    def update_cache(self, url: Text, status_code: Optional[int], report_hash: Optional[Text]) -> None:
        """ Add url/hash to cache """
        cursor = self.cache.cursor()

        # Check if url exists
        res = cursor.execute("SELECT * FROM report_hash WHERE url = ?", [url.strip()]).fetchall()

        if res:
            info("Update cache {}, {}, {}".format(url, status_code, report_hash))
            cursor.execute(
                "UPDATE report_hash set status_code = ?, sha256 = ?, added = ? where url = ?",
                [status_code, report_hash, int(time.time()), url])
        else:
            info("Insert cache {} -> {}".format(url, report_hash))
            cursor.execute("INSERT INTO report_hash VALUES (?,?,?,?)", [url, status_code, report_hash, time.time()])

        self.cache.commit()

    def url_sha256(self, url: Text) -> UrlContentHash:
        "Retrieve URL and return sha256 of content. Returns None if request fails."

        sha256 = None
        status_code = None

        try:
            req = requests.get(url, **self.requests_common_kwargs)
            status_code = req.status_code
            if req.status_code == 200:
                sha256 = hashlib.sha256(req.content).hexdigest()
            else:
                info("Failed downloading {}: {}".format(url, req.status_code))
        except requests.exceptions.ReadTimeout:
            info("Timeout downloading {}".format(url))
        except Exception as err:  # pylint: disable=broad-except
            info("Unknown exception downloading {}: {}".format(url, err))

        return UrlContentHash(status_code, sha256)

    def query_download_update(self, url: Text) -> Optional[Text]:
        """

        Query cache for url. If not found in url, or cache entry is old, download content

        Returns sha256 of content, or None if unable to download

        """

        cache = self.query_cache(url)

        now = datetime.datetime.now()

        if cache.report_hash and cache.added > now - datetime.timedelta(days=7):
            info("URL in cache (with hash found): {}, {}, {}, {}".format(url, cache.report_hash, cache.status_code, time.time()))
        elif (not cache.report_hash) and (cache.status_code == 404) and (cache.added > now - datetime.timedelta(days=7)):
            info("URL in cache (last status was 404): {}, {}, {}, {}".format(url, cache.report_hash, cache.status_code, time.time()))
        elif (not cache.report_hash) and (cache.added > now - datetime.timedelta(days=2)):
            # status_code != 404 - attempt new download
            info("URL in cache (Unknown status): {}, {}, {}, {}".format(url, cache.report_hash, cache.status_code, time.time()))
        else:
            query = self.url_sha256(url)
            self.update_cache(url, query.status_code, query.sha256)
            return query.sha256

        return cache.report_hash



#!/usr/bin/env python3

"""General ACT backend uploader. Reads facts as JSON
from the stdin, uploading accordingly"""

import argparse
import json
import sys
import time
import traceback
from logging import error, warning

import act.api

from act.workers.libs import worker


def parseargs() -> argparse.ArgumentParser:
    """ Parse arguments """
    parser = worker.parseargs('Generic uploader')
    parser.add_argument('--timing', action="store_true",
                        help="Add timing operations at warn level")

    return parser


def main(actapi: act.api.Act, timing: bool = False) -> None:
    """Process stdin, parse each separat line as a JSON structure and
    register a fact based on the structure. The form of input should
    be the on the form accepted by the ACT Rest API fact API."""

    handle_fact_time = []
    origins = set()

    for line in sys.stdin:
        try:
            data = json.loads(line)
        except json.decoder.JSONDecodeError:
            error("Error decoding json: %s" % line)
            continue

        fact = actapi.fact(**data)
        try:
            started = time.time()
            act.api.helpers.handle_fact(fact)

            if fact.origin:
                origins.add(fact.origin.name)

            time_spent = time.time() - started
            handle_fact_time.append(time_spent)
            if timing:
                warning("Handle fact time: %s", round(time_spent, 2))
        except act.api.base.ValidationError as err:
            warning("ValidationError while storing objects: %s" % err)
        except act.api.base.ResponseError as err:
            error("ResponseError while storing objects: %s" % err)
            sys.exit(1)

    if timing and handle_fact_time:
        warning(
            "Total time (count:%s,total:%s,mean:%s,min:%s,max:%s,origins:%s)",
            len(handle_fact_time),
            round(sum(handle_fact_time), 2),
            round(sum(handle_fact_time)/len(handle_fact_time), 2),
            round(min(handle_fact_time), 2),
            round(max(handle_fact_time), 2),
            "+".join(origins)
        )


def main_log_error() -> None:
    "Call main() and log all exceptions as errors"
    try:
        # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
        # (or replace .config with $XDG_CONFIG_DIR if set)
        args = worker.handle_args(parseargs())
        actapi = worker.init_act(args)

        main(actapi, args.timing)
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == '__main__':
    main_log_error()



""" Common functions towards mnemonic API """

from logging import debug
from typing import Any, Dict, Generator, Optional, Text

import requests

from . import worker

# Mapping of message templates provided in 402/503 errors from backend to
# Exceptions that will be raised

STATUS_CODE_TEMPLATES = {
    402: {
        "resource.limit.exceeded": lambda msg: worker.ResourceLimitExceeded(
            "{message}".format(**msg)),
    },

    503: {
        "service.timeout": lambda msg: worker.ServiceTimeout(
            "{message} ({field}={parameter})".format(**msg)),
    }
}


def status_code_handler(req: requests.models.Response, res: Dict) -> None:
    """

    Status Code Handler for non-200 status codes
    Raise exceptions on combinations of status code and message

    Arguments:

    req: requests Response object
    res: requests Response content json parsed

    """

    if req.status_code in (402, 503):
        # Example output on Timeout:
        # {"responseCode":503,"limit":0,"offset":0,"count":0,"metaData":{},"messages":[{"message":null,"messageTemplate":null,"type":"ACTION_ERROR","field":null,"parameter":null,"timestamp":1606137003885},{"message":"Request timed out, service may be overloaded or unavailable","messageTemplate":"service.timeout","type":"ACTION_ERROR","field":null,"parameter":null,"timestamp":1606137003993}],"data":null,"size":0}

        for msg in res["messages"]:
            msg_template = msg.get("messageTemplate")
            if msg_template in STATUS_CODE_TEMPLATES.get(req.status_code, {}):
                raise STATUS_CODE_TEMPLATES[req.status_code][msg_template](msg)

    raise worker.UnknownResult("Unknown error: {}, {}".format(req, req.content))


def batch_query(
        method: Text,
        url: Text,
        headers: Optional[Dict] = None,
        timeout: int = 299,
        json_params: Optional[Dict] = None,
        proxy_string: Optional[Text] = None,
        batch_size: int = 1000,
        limit: int = 0) -> Generator[Dict[Text, Any], None, None]:
    """ Execute query until we have all results """

    offset = 0
    count = 0

    proxies = {
        'http': proxy_string,
        'https': proxy_string
    }

    if limit and batch_size > limit:
        batch_size = limit

    options = {
        "headers": headers,
        "verify": False,
        "timeout": timeout,
        "proxies": proxies,
        "params": {}
    }

    while True:  # do - while offset < count
        # ARGUS uses offset in body for POST requests and request parameters for GET requests

        if method == "POST" and json_params:
            json_params["offset"] = offset
            json_params["limit"] = batch_size
        elif method == "GET":
            options["params"]["offset"] = offset  # type: ignore
            options["params"]["limit"] = batch_size

        debug("Executing search: {}, json={}, options={}".format(url, json_params, options))
        req = requests.request(method, url, json=json_params, **options)  # type:ignore

        try:
            res = req.json()
        except ValueError:
            raise worker.UnknownResult("Illegal JSON, {}, {}".format(req, req.content))

        if req.status_code != 200:
            status_code_handler(req, res)

        data = res["data"]
        count = res.get("count", 0)

        if type(data) != type([]):
            yield data
            break

        yield from data

        debug("count={}".format(count))

        offset += len(data)

        # if we have defined a limit, stop processing on reaching limit
        if limit and offset >= limit:
            break

        if offset >= count:
            break


def single_query(
        method: Text,
        url: Text,
        headers: Optional[Dict] = None,
        timeout: int = 299,
        json_params: Optional[Dict] = None,
        proxy_string: Optional[Text] = None) -> Dict[Text, Any]:
    """ Execute query for single result, returns result """

    try:
        for res in batch_query(method, url, headers, timeout, json_params, proxy_string):
            return res
    except worker.UnknownResult as e:
        if not str(e).startswith("Unknown error: 404"):
            raise

    raise worker.NoResult()



#!/usr/bin/env python3
'''Alienvault OTX Worker for the ACT Project

Copyright 2018 the ACT project <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

'''

import argparse
import datetime
import hashlib
import logging
import os
import sys
import traceback
import urllib.parse
from typing import Any, Dict, Generator, Optional, Text

import act
import act.api
import requests
from act.workers.libs import worker

WORKER_NAME = 'alienvault-otx'
VERSION = 0.1

OTX_ACT_TYPE_MAPPING: Dict[Text, Text] = {
    'CVE': 'vulnerability',
    'FileHash-MD5': 'hash',
    'FileHash-SHA1': 'hash',
    'FileHash-SHA256': 'hash',
    'FilePath': 'uri',
    'IPv4': 'ipv4',
    'IPv6': 'ipv6',
    'Mutex': 'mutex',
    'URI': 'path',
    'URL': 'uri',
    'domain': 'fqdn',
    'email': 'uri',
    'hostname': 'fqdn',
}


class AuthorizationError(Exception):
    """AutorizationError raised if API returns status code 403"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


class AlienvaultOTXAPI:
    """Class for interacting with Alienvault OTX API"""

    def __init__(self, args: argparse.Namespace) -> None:
        """Initialisation method"""
        self.args = args

    def __api_request(self, api_url: Text) -> Any:
        """Method for sending API requests to the Alienvault OTX API"""

        # add necessary headers
        headers: Dict[Text, Text] = {
            'Content-Type': 'application/json',
            'X-OTX-API-KEY': self.args.otx_apikey,
        }

        # check if we should use a proxy
        proxy: Dict[Text, Text] = {}
        if 'proxy_string' in self.args:
            proxy = {
                'http': self.args.proxy_string,
                'https': self.args.proxy_string,
            }

        # make api request
        request = requests.get(
            api_url,
            headers=headers,
            proxies=proxy,
        )

        # check for 403 status code
        if request.status_code == 403:
            raise AuthorizationError('Authentication required')

        # will raise exception if no json is returned which is how the api behaves
        # if you specify a wrong url
        return request.json()

    def get_subscribed(self) -> Generator[Dict[Any, Any], None, None]:
        """Method for getting events from pulses you are subscribed to"""

        # get last updated timestamp
        last_updated = self.last_retrived()

        # generate full api url
        if last_updated:
            api_url = urllib.parse.urljoin(
                self.args.otx_baseurl,
                f'v1/pulses/subscribed?modified_since={last_updated}'
            )
        else:
            api_url = urllib.parse.urljoin(self.args.otx_baseurl, 'v1/pulses/subscribed')

        # do api request
        data = self.__api_request(api_url)

        # iterate through pages with results
        while True:
            for entry in data['results']:
                yield entry

            if 'next' in data and data['next']:
                data = self.__api_request(data['next'])
            else:
                break

        # update last retrived timestamp
        self.last_retrived(update=True)

    def last_retrived(self, update: bool = False) -> Optional[Text]:
        """Method for handling last retrived timestamp"""

        # if update is set, just write current timestamp to filepath
        if update:
            with open(self.args.otx_path_lastretrived, 'w') as file_handle:
                timestamp = datetime.datetime.utcnow().isoformat()
                file_handle.write(timestamp)

        # update is not set, try to return previous timestamp
        else:
            # file does not exist, return None
            if not os.path.isfile(self.args.otx_path_lastretrived):
                return None

            # try to read file, parse timestamp, and return it
            with open(self.args.otx_path_lastretrived, 'r') as file_handle:
                try:
                    timestamp = str(datetime.datetime.strptime(file_handle.read(), '%Y-%m-%dT%H:%M:%S.%f'))
                    return timestamp
                except ValueError:
                    return None

        # fallback
        return None


class ConfigurationError(Exception):
    """Raised when invalid option for events to import is used"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


def parseargs() -> argparse.ArgumentParser:
    """Extract command lines argument"""

    parser = worker.parseargs('ACT Alienvault OTX Import Client v{}'.format(VERSION))
    parser.add_argument('--config_path', metavar='PATH',
                        help='Path to reputation configuration files')
    parser.add_argument('--otx-baseurl', default="https://otx.alienvault.com/api/",
                        help="Alienvault OTX API host")
    parser.add_argument('--otx-apikey', metavar='KEY',
                        help='Alienvault OTX API key')
    parser.add_argument('--otx-path-lastretrived', metavar='FILEPATH',
                        help='Path to store last retrival timestamp')
    return parser


def strip_special_chars(msg: Text) -> Text:
    """Strips special chars from a string """
    for char in ['\r\n', '\n', '\r']:
        msg = msg.replace(char, ' ')

    return msg


def handle_facts(actapi: act.api.helpers.Act, event: Dict[Any, Any]) -> None:
    """Generates a list of json facts based on a given ioc"""

    # generate report name - sha256
    report_id: Text = hashlib.sha256(
            f'alienvault-otx-{event["id"]}-{event["modified"]}'.encode('utf-8')
    ).hexdigest()

    # add a name fact to the report
    if 'name' in event and event['name']:
        name_fact = actapi.fact('name', strip_special_chars(event['name']))
        name_fact.source('report', report_id)
        act.api.helpers.handle_fact(name_fact, output_format='json')

    # iterate over all indicators
    for ioc in event['indicators']:

        if not ioc['type'] in OTX_ACT_TYPE_MAPPING:
            continue

        # identify act fact type
        act_type: Text = OTX_ACT_TYPE_MAPPING[ioc['type']]

        # add schemas for certain ioc types
        if ioc['type'] == 'email':
            ioc['indicator'] = f'email://{ioc["indicator"]}'
        elif ioc['type'] == 'FilePath':
            ioc['indicator'] = f'file://{ioc["indicator"]}'

        # special treatment for uri's
        if act_type in ['uri']:
            act.api.helpers.handle_uri(actapi, ioc['indicator'], output_format='json')
        else:
            # create fact
            if 'description' in ioc and ioc['description']:
                fact = actapi.fact('mentions', comment=strip_special_chars(ioc['description']))
            else:
                fact = actapi.fact('mentions')

            fact.source('report', report_id)

            # some facts should not be lower cased
            # uri is handled above, but added for safety
            if act_type in ['mutex', 'uri']:
                fact.destination(act_type, ioc['indicator'])
            else:
                fact.destination(act_type, ioc['indicator'].lower())

            # output fact
            act.api.helpers.handle_fact(fact, output_format='json')

            # if fact type is a hash, we also have to create a fact which tells us what kind
            # of hash it represents
            if act_type in ['hash']:
                hash_type: Text = ioc['type'].split('-')[-1].lower()
                hash_fact = actapi.fact('category', hash_type)
                hash_fact.source(act_type, ioc['indicator'])
                act.api.helpers.handle_fact(hash_fact, output_format='json')


def main() -> None:
    """Main function"""
    # Look for default ini file in "/etc/actworkers.ini" and
    # ~/config/actworkers/actworkers.ini (or replace .config with
    # $XDG_CONFIG_DIR if set)
    args = worker.handle_args(parseargs())

    # setup logging
    act.api.utils.setup_logging(args.loglevel, prefix='act-alienvault-otx')

    # check necessary configuration items for errors
    for item in ['config_path', 'otx_baseurl', 'otx_apikey', 'otx_path_lastretrived']:
        if item not in args:
            raise ConfigurationError('Missing configuration item {}'.format(item))

    # initialise act api
    actapi = worker.init_act(args)

    # initialise otx api
    otxapi = AlienvaultOTXAPI(args)

    # create facts for indicators in each event
    for event in otxapi.get_subscribed():
        handle_facts(actapi, event)


def main_log_error() -> None:
    "Main function wrapper. Log all exceptions to error"
    pid_file = '/tmp/act-alienvault-otx.pid'
    if os.path.isfile(pid_file):
        logging.error('Instance already running')
        sys.exit(1)

    try:
        with open(pid_file, 'w') as pid_fh:
            pid_fh.write(str(os.getpid()))
        main()
    except Exception:
        logging.error('Unhandled exception: %s', traceback.format_exc())
        raise
    finally:
        os.unlink(pid_file)


if __name__ == '__main__':
    main_log_error()



#!/usr/bin/env python3

"""url unshortener worker for the ACT platform

Copyright 2018 the ACT project <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
"""


from logging import error, warning

import urllib3

from act.workers.libs import worker
from typing import Text, List, Dict
from urllib.parse import urlparse
import act
import act.api
import argparse
import requests
import sys
import traceback

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

MAX_RECURSIVE = 10  # max number of redirects to attempt to follow (failsafe)


def check_redirect(url: Text, url_shorteners: List[Text], user_agent: Text,
                   proxies: Dict[Text, Text], timeout: int = 30) -> Text:
    """Take a url. Attempt to make it a http:// url and check if it is to one
    of the known url shortening services. If it is.. find the first redirect"""

    headers = {'User-agent': user_agent}

    org_url = url

    p = urlparse(url)
    if p.scheme == '':
        url = "http://{}".format(url)
    p = urlparse(url)

    if p.hostname not in url_shorteners:
        return org_url

    r = requests.get(url, allow_redirects=False, timeout=timeout, headers=headers, proxies=proxies)
    if r.is_redirect:
        return str(r.next.url)  # type: ignore

    return org_url


def process(api: act.api.Act, shorteners: List[Text], user_agent: Text,
            proxies: Dict[Text, Text],
            output_format: Text = "json") -> None:
    """Read queries from stdin, resolve each one through passivedns printing
    generic_uploader data to stdout"""

    for query in sys.stdin:
        query = query.strip()
        if not query:
            continue

        n = 0
        while True:
            redirect = check_redirect(query, shorteners, user_agent, proxies)
            if redirect == query or n > MAX_RECURSIVE:
                break
            n += 1

            try:
                act.api.helpers.handle_uri(api, query, output_format=output_format)
            except act.api.base.ValidationError as err:
                warning("Unable to add {0} [{1}]".format(query, err))
                break

            try:
                act.api.helpers.handle_uri(api, redirect, output_format=output_format)
            except act.api.base.ValidationError as err:
                warning("Unable to add {0} [{1}]".format(redirect, err))
                break

            act.api.helpers.handle_fact(
                api.fact("redirectsTo")
                .source("uri", query)
                .destination("uri", redirect), output_format=output_format)

            query = redirect


def parseargs() -> argparse.Namespace:
    parser = worker.parseargs("URL unshortener worker")

    parser.add_argument("--url-shorteners", dest="url_shorteners",
                        default="adf.ly, bit.ly, bitly.com, cc.uz, evassmat.com, goo.gl, is.gd, lnkd.in, smarturl.it, www.t2m.io, tiny.cc, tinyurl.com, x.co",
                        help="Comma separated list of shortener-domains")
    parser.add_argument("--user-agent", dest="user_agent",
                        default="Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.169 Safari/537.36",
                        help="User-agent to present to the redirect services")

    return worker.handle_args(parser)


def main() -> None:
    """Main function"""
    # Look for default ini file in "/etc/actworkers.ini" and
    # ~/config/actworkers/actworkers.ini (or replace .config with
    # $XDG_CONFIG_DIR if set)
    args = parseargs()

    try:
        shorteners = [x.strip() for x in args.url_shorteners.split(",")]
    except AttributeError:
        worker.fatal("Empty list of shorteners?")

    actapi = worker.init_act(args)

    proxies = {
        'http': args.proxy_string,
        'https': args.proxy_string
    } if args.proxy_string else None

    process(actapi, shorteners, args.user_agent, proxies, args.output_format)


def main_log_error() -> None:
    "Main function wrapper. Log all exceptions to error"
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == '__main__':
    main_log_error()



#!/usr/bin/env python3

"""Origin client. Manage origins """

import argparse
import re
import sys
import traceback
from logging import error
from typing import Text

import act.api
from act.workers.libs import worker


def parseargs() -> argparse.ArgumentParser:
    """ Parse arguments """
    parser = worker.parseargs('Manage origins')
    parser.add_argument("--list", action="store_true", help="List origins")
    parser.add_argument("--add", action="store_true", help="List origins")
    parser.add_argument("--delete", action="store_true", help="List origins")

    # Trust is converted to float before sending a request to the platform
    # and since this value can come from an ini file (where it will be a string)
    # We keep the value as a string here
    parser.add_argument("--default-trust", default="0.8", help="Default trust")

    return parser


def fatal(msg: Text, exit_code: int = 1) -> None:
    "Output message as error and exit"
    error(msg)
    sys.exit(exit_code)


def add_origin(actapi: act.api.Act, default_trust: Text) -> None:
    sys.stdout.write("Origin name: ")
    name = input()
    sys.stdout.write("Origin description: ")
    description = input()
    sys.stdout.write("Origin trust (float 0.0-1.0. Default=0.8): ")
    trust = input()
    sys.stdout.write("Origin organization (UUID): ")
    organization = input()

    if not trust:
        trust = default_trust

    try:
        trust = float(trust)
    except ValueError:
        fatal("Unable to convert {} to float".format(trust))

    if not (trust >= 0.0 and trust <= 1.0):
        fatal("Trust must be between 0.0 and 1.0")

    params = {
        "name": name,
        "description": description,
        "trust": trust,
    }

    if organization:
        if not re.search(act.api.re.UUID_MATCH, organization):
            fatal("Organization must be a valid UUID")

        params["organization"] = organization

    origin = actapi.origin(**params)
    origin.add()

    print("Origin added:")
    print(origin)


def origin_handler(actapi: act.api.Act, args: argparse.Namespace) -> None:
    "handle origins"

    try:
        if args.list:
            for origin in actapi.get_origins():
                print(origin)

        if args.add:
            add_origin(actapi, default_trust=args.default_trust)

        if args.delete:
            actapi.api_delete("v1/origin/uuid/{}".format(args.origin_id))

            print("Origin deleted: {}".format(args.origin_id))

    except act.api.base.ResponseError as err:
        error("ResponseError while connecting to platform: %s" % err)


def main() -> None:
    "main function"
    try:
        # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
        # (or replace .config with $XDG_CONFIG_DIR if set)
        args = worker.handle_args(parseargs())

        if not (args.act_baseurl and args.user_id):
            fatal("--act-baseurl and --user-id must be specified")

        if not (args.list or args.add or args.delete):
            fatal("Specify either --list, --add or --delete")

        if (args.delete) and not (args.origin_id):
            fatal("Specify --origin-id to delete an origin")

        actapi = worker.init_act(args)
        origin_handler(actapi, args)
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"Search for objects and traverse graph for all results"

import argparse
import concurrent.futures
import configparser
import datetime
import inspect
import os
import re
import sys
from logging import error, info, warning
from typing import Any, Dict, Text

import dateparser

import act.api
from act.workers.libs import worker

# Inspect object search function to get list of valid arguments and their defaults
VALID_SEARCH_OPTIONS: Dict[str, Any] = {
    arg: parameter.default
    for arg, parameter in inspect.signature(act.api.Act.object_search).parameters.items()
    if arg != "self"
}

SEARCH_OPTIONS_FUNC: Dict = {
    "before": lambda ts: dateparser.parse(ts).strftime(act.api.ACT_TIME_FORMAT),
    "after": lambda ts: dateparser.parse(ts).strftime(act.api.ACT_TIME_FORMAT),
}


def parseargs() -> argparse.ArgumentParser:
    """ Parse arguments """
    parser = worker.parseargs('Search Graph')
    parser.add_argument('--search-jobs', help="Search jobs (ini-file)")
    parser.add_argument(
        '--output-path',
        help="Output-path for result files (default=.)",
        default=".")
    parser.add_argument(
        '--workers',
        type=int,
        default=4,
        help="Number of parallel workers for graph search")

    return parser


def parse_search_jobs(config_filename: Text) -> Dict:
    "Parse config with search jobs and return dictionary of section name and options"
    config = configparser.ConfigParser()
    config.read(config_filename)
    return {section: dict(config.items(section)) for section in config.sections()}


def traverse(api: act.api.Act, name: Text, weburl: Text, obj: act.api.obj.Object, query: Text, minfacts: int) -> Text:
    output = ""
    try:
        # Replace all whitespace that is not within quotes
        # https://stackoverflow.com/questions/9577930/regular-expression-to-select-all-whitespace-that-isnt-in-quotes

        query_url = re.sub(r'\s+(?=([^"]*"[^"]*")*[^"]*$)', '', query)
        info("{} -> {}".format(obj, query_url))
        res = api.object(obj.type, obj.value).traverse(query)

        facts = {str(obj) for obj in res if isinstance(obj, act.api.fact.Fact)}
        if len(facts) >= minfacts:
            output += """[{}:{}]\n{}/graph-query/{}/{}/{}\n""".format(
                name, obj, weburl, obj.type.name, obj.value, query_url)

            # String representation of all facts (joined with newlines)
            output += "\n".join(facts)

    except act.api.base.ResponseError as e:
        error("{}, {}, {}".format(e, obj, query))

    return output


def process(actapi: act.api.Act, output_path: Text, name: Text, options: Dict, workers: int) -> None:
    if "query" not in options:
        error("No query specified in {}, skipping".format(name))
        return

    # Extract query as separate option
    query = options["query"]
    blacklist = options.get("blacklist", "").split(",")
    object_value_re = options.get("object_value_re", r'.*')
    weburl = options.get("weburl", actapi.config.act_baseurl)
    minfacts = int(options.get("minfacts", 1))

    filename = os.path.join(
        output_path,
        "{}-{}.result".format(datetime.datetime.now().strftime("%Y-%m-%d-%H-%m"), name))

    info("Result will be written to {}".format(filename))

    with open(filename, "w") as f:

        # Construct search options from ini file
        search_options = {}

        for key, value in options.items():
            if key in ("query", "blacklist", "object_value_re", "weburl", "minfacts"):
                # Query/blacklist/object_value_re is not part of object search options
                # but passed on to traverse
                continue

            if key not in VALID_SEARCH_OPTIONS:
                error("Illegal query option ({}) specified in {}, skipping".format(key, name))
                return

            if key in SEARCH_OPTIONS_FUNC:
                search_options[key] = SEARCH_OPTIONS_FUNC[key](value)
            elif VALID_SEARCH_OPTIONS[key] == []:
                # Split values on comma (except escaped commas)
                search_options[key] = [value.strip() for value in re.split(r'(?<!\\),', value) if value]
            else:
                search_options[key] = value

        info("Search options: {}".format(search_options))

        try:
            objects = actapi.object_search(**search_options)
        except Exception:
            error("Fatal: object search exception: {}".format(search_options), exc_info=True)
            return

        if objects.size != objects.count:
            warning("Received only {}/{} objects".format(objects.size, objects.count))

        info("Received {} objects".format(objects.size))

        with concurrent.futures.ProcessPoolExecutor(max_workers=workers) as executor:

            future_traverse = {
                executor.submit(traverse, actapi, name, weburl, obj, query, minfacts): str(obj)
                for obj in objects
                if (str(obj) not in blacklist) and re.search(object_value_re, obj.value)
            }

            for idx, future in enumerate(concurrent.futures.as_completed(future_traverse)):
                obj = future_traverse[future]
                try:
                    data = future.result()
                except Exception:
                    error("traverse exception: {}".format(obj), exc_info=True)
                    continue

                info("Progress: {}/{}".format(idx + 1, objects.size))

                if data:
                    f.write(data + "\n\n")
                    f.flush()


def main_log_error() -> None:
    "Main function. Log all exceptions to error"
    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)

    args = worker.handle_args(parseargs())
    actapi = worker.init_act(args)

    if not (args.act_baseurl and args.user_id):
        sys.stderr.write("Must specify --baseurl and --user-id\n")
        sys.exit(1)

    if not os.path.isfile(args.search_jobs):
        sys.stderr.write("File not found: {}\n".format(args.search_jobs))
        sys.exit(2)

    if not args.search_jobs:
        sys.stderr.write("Must specify config file with search jobs\n")
        sys.exit(2)

    try:
        for name, options in parse_search_jobs(args.search_jobs).items():
            process(actapi, args.output_path, name, options, args.workers)

    except Exception:
        error("Unhandled exception", exc_info=True)
        raise


if __name__ == '__main__':
    main_log_error()



#!/usr/bin/env python3

"""ip->asn history worker for the ACT platform

Copyright 2019 mnemonic AS <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
"""


from logging import error

import sys
import traceback
import act
import act.api
import requests
from act.workers.libs import worker
from typing import Text, Optional, List, Tuple


WORKER_NAME = "ip-asn-history"


def lookup_ip(ip: Text, proxy: Optional[Text] = None) -> List[Tuple[Text, Text]]:
    """Lookup historc ASN registration for an IP against the CIRCL online ASN database"""

    proxies = {
        'http': proxy,
        'https': proxy} if proxy else None

    r = requests.get('https://bgpranking-ng.circl.lu/ipasn_history/?ip={}'.format(ip), proxies=proxies)

    data = r.json()

    return [(x['asn'], x['prefix']) for x in data['response'].values()]


def process(api: act.api.Act, proxy=None, output_format: Text = "json") -> None:
    """Read queries from stdin"""

    for query in sys.stdin:
        query = query.strip()

        if not query:
            continue

        for asn, network in lookup_ip(query, proxy):
            act.api.helpers.handle_fact(
                api.fact('memberOf')
                .source('ipv4', query)
                .destination('ipv4Network', network), output_format=output_format)

            act.api.helpers.handle_fact(
                api.fact('memberOf')
                .source('ipv4Network', network)
                .destination('asn', asn), output_format=output_format)


def main() -> None:
    """Main function"""
    # Look for default ini file in "/etc/actworkers.ini" and
    # ~/config/actworkers/actworkers.ini (or replace .config with
    # $XDG_CONFIG_DIR if set)
    args = worker.handle_args(worker.parseargs(WORKER_NAME))
    actapi = worker.init_act(args)

    process(actapi, args.proxy_string, args.output_format)


def main_log_error() -> None:
    "Main function wrapper. Log all exceptions to error"
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == '__main__':
    main_log_error()



"""Common worker library"""

import argparse
import inspect
import os
import re
import smtplib
import socket
import sys
from email.mime.text import MIMEText
from logging import debug, error, warning
from typing import Any, Optional, Text, Dict, cast

import caep

import requests
import urllib3

import act.api

CONFIG_ID = "actworkers"
CONFIG_NAME = "actworkers.ini"


class UnknownResult(Exception):
    """UnknownResult is used in API request (not 200 result)"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


class NoResult(Exception):
    """NoResult is used in API request (no data returned)"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


class UnknownFormat(Exception):
    """UnknownFormat is used on unknown parsing formats"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


class ResourceLimitExceeded(Exception):
    """ Resource Limits Exceeded"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


class ServiceTimeout(Exception):
    """ Internal service timeouts """

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)



def parseargs(description: str) -> argparse.ArgumentParser:
    """ Parse arguments """
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="{} ({})".format(description, worker_name()), epilog="""

  --config INI_FILE     Override default locations of ini file

    Arguments can be specified in ini-files, environment variables and
    as command line arguments, and will be parsed in that order.

    By default, workers will look for an ini file in /etc/{1}
    and ~/.config/{0}/{1} (or in $XDG_CONFIG_DIR if
    specified).

    Each worker will read the confiuration from the "DEFAULT" section in the
    ini file, and in it's own section (in that order).

    It is also possible to use environment variables for configuration.
    Workers will look for environment variables for all arguments with
    the argument name in uppercase and "-" replaced with "_".

    E.g. set the CERT_FILE environment variable to configure the
    --cert-file option.

""".format(CONFIG_ID, CONFIG_NAME), formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('--http-timeout', dest='http_timeout', type=int,
                        default=120, help="Timeout")
    parser.add_argument('--proxy-string', dest='proxy_string', help="Proxy to use for external queries")
    parser.add_argument('--proxy-platform', dest='proxy_platform', action="store_true", help="Use proxy-string towards the ACT platform")
    parser.add_argument('--cert-file', dest='cert_file', help="Cerfiticate to add if you are behind a SSL/TLS interception proxy.")
    parser.add_argument('--user-id', dest='user_id',
                        help="User ID")
    parser.add_argument('--act-baseurl', dest='act_baseurl',
                        help='ACT API URI')
    parser.add_argument("--logfile", dest="logfile",
                        help="Log to file (default = stdout)")
    parser.add_argument("--loglevel", dest="loglevel", default="info",
                        help="Loglevel (default = info)")
    parser.add_argument("--output-format", dest="output_format", choices=["str", "json"], default="json",
                        help="Output format for fact (default=json)")
    parser.add_argument('--access-mode', default=act.api.DEFAULT_ACCESS_MODE,
                        choices=act.api.ACCESS_MODES,
                        help="Specify default access mode used for all facts.")
    parser.add_argument('--organization', help="Specify default organization applied to all facts.")
    parser.add_argument('--http-header', help="Comma separated list of HTTP headers, e.g. 'HeaderA: val1, HeaderB:comma\,val2")
    parser.add_argument('--http-user', dest='http_user', help="ACT HTTP Basic Auth user")
    parser.add_argument('--http-password', dest='http_password', help="ACT HTTP Basic Auth password")
    parser.add_argument('--disabled', dest='disabled', action="store_true", help="Worker is disabled (exit immediately)")
    parser.add_argument('--origin-name', dest='origin_name', help="Origin name. This name must be defined in the platform")
    parser.add_argument('--origin-id', dest='origin_id', help="Origin id. This must be the UUID of the origin in the platform")
    return parser


def __mod_name(stack: inspect.FrameInfo) -> Text:
    """ Return name of module from a stack ("_" is replaced by "-") """
    mod = inspect.getmodule(stack[0])
    return os.path.basename(mod.__file__).replace(".py", "").replace("_", "-")


def worker_name() -> Text:
    """ Return first external module that called this function, directly, or indirectly """

    modules = [__mod_name(stack) for stack in inspect.stack() if __mod_name(stack)]
    return [name for name in modules if name != modules[0]][0]


def handle_args(parser: argparse.ArgumentParser) -> argparse.Namespace:
    """ Wrapper for caep.handle_args where we set config_id and config_name """
    args = caep.handle_args(parser, CONFIG_ID, CONFIG_NAME, worker_name())

    if args.http_header:
        # Convert comma separated list of http headers to dictionary
        headers = {}

        # Split on comma, unless they are escaped
        for header in re.split(r'(?<!\\),', args.http_header):
            if ":" not in header:
                raise act.api.base.ArgumentError(f"No ':' in header, http header: {header}")
            header_key, header_val = header.split(":", 1)
            header_key = header_key.strip().replace("\\,", ",")
            header_val = header_val.strip().replace("\\,", ",")
            headers[header_key] = header_val
        args.http_header = headers

    return cast(argparse.Namespace, args)


def init_act(args: argparse.Namespace) -> act.api.Act:
    """ Initialize act api from arguments """
    requests_kwargs: Dict[Text, Any] = {}

    if args.http_header:
        requests_kwargs["headers"] = args.http_header

    if args.http_user:
        requests_kwargs["auth"] = (args.http_user, args.http_password)

    if args.proxy_string and args.proxy_platform:
        requests_kwargs["proxies"] = {
            "http": args.proxy_string,
            "https": args.proxy_string
        }

    if args.cert_file:
        requests_kwargs["verify"] = args.cert_file

    api = act.api.Act(
        args.act_baseurl,
        args.user_id,
        args.loglevel,
        args.logfile,
        worker_name(),
        requests_common_kwargs=requests_kwargs,
        origin_name=args.origin_name,
        origin_id=args.origin_id,
        access_mode=args.access_mode,
        organization=args.organization
    )

    if args.http_header:
        # Debug output of HTTP headers (must wait until act.api.Act() is initialized
        # so we have setup logging)
        debug("HTTP headers: %s", args.http_header)

    # This check is done here to make sure logging is set up
    if args.disabled:
        warning("Worker is disabled")
        sys.exit(0)

    return api


def fatal(message: Text, exit_code: int = 1) -> None:
    "Send error to error() and stderr() and exit with exit_code"
    sys.stderr.write(message.strip() + "\n")
    error(message.strip())
    sys.exit(exit_code)


def fetch_json(url: str, proxy_string: Optional[str], timeout: int = 60, verify_https: bool = False) -> Any:
    """Fetch remote URL as JSON
    url (string):                    URL to fetch
    proxy_string (string, optional): Optional proxy string on format host:port
    timeout (int, optional):         Timeout value for query (default=60 seconds)
    """

    proxies = {
        'http': proxy_string,
        'https': proxy_string
    }

    options = {
        "verify": verify_https,
        "timeout": timeout,
        "proxies": proxies,
        "params": {}
    }

    if not verify_https:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    try:
        req = requests.get(url, **options)
    except (urllib3.exceptions.ReadTimeoutError,
            requests.exceptions.ReadTimeout,
            socket.timeout) as err:
        error("Timeout ({0.__class__.__name__}), query: {1}".format(err, req.url))

    if not req.status_code == 200:
        errmsg = "status_code: {0.status_code}: {0.content}"
        raise UnknownResult(errmsg.format(req))

    return req.json()


def sendmail(smtphost: str, sender: str, recipient: str, subject: str, body: str) -> None:
    """Send email"""

    msg = MIMEText(body, "plain", "utf-8")
    msg['Subject'] = subject
    msg['From'] = sender
    msg['To'] = recipient
    s = smtplib.SMTP(smtphost)
    s.sendmail(sender, [recipient], msg.as_string())
    s.quit()
